import connect from '../connect';
import { getMapping } from '../connect';
import { Machine } from '../../';

describe('Given the connect helper', function () {
  beforeEach(() => {
    Machine.flush();
  });
  describe('when using connect', function () {
    it('should allow mapping to machines', function (done) {
      Machine.create('A', {
        state: { name: 'idle' },
        transitions: {
          idle: { run: 'running' },
          running: { stop: 'idle' }
        }
      });
      Machine.create('B', {
        state: { name: 'waiting' },
        transitions: {
          waiting: { fetch: 'fetching' },
          fetching: { done: 'waiting' }
        }
      });
      connect()
        .with('A', 'B')
        .map((A, B) => {

          expect(A.state.name).to.equal('idle');
          expect(B.state.name).to.equal('waiting');
          done();
        });
    });
    it('should add only one middleware', function () {
      const mappingA = sinon.spy();
      const mappingB = sinon.spy();
      const machine = Machine.create('A', {
        state: { name: 'idle' },
        transitions: {
          idle: { run: 'running' },
          running: { stop: 'idle' }
        }
      });

      connect().with('A').map(mappingA);
      connect().with('A').map(mappingB);
      connect().with(machine).map(mappingB);

      expect(Machine.middlewares.length).to.be.equal(1);
    });
    describe('and when we update the state of the mapped machine/s', function () {
      it('should fire the mapping function', function () {
        const mapping = sinon.spy();
        const machine = Machine.create('A', {
          state: { name: 'idle' },
          transitions: {
            idle: { run: 'running' },
            running: { stop: 'idle' }
          }
        });

        connect().with('A').map(A => {
          mapping(A.state.name);
        });

        machine.run();
        machine.stop();

        expect(mapping).to.be.calledThrice;
        expect(mapping.firstCall).to.be.calledWith(sinon.match('idle'));
        expect(mapping.secondCall).to.be.calledWith(sinon.match('running'));
        expect(mapping.thirdCall).to.be.calledWith(sinon.match('idle'));
      });
    });
    describe('and we use `mapOnce`', function () {
      it('should fire the mapping only once', function () {
        const mapping = sinon.spy();
        const machine = Machine.create('A', {
          state: { name: 'idle' },
          transitions: {
            idle: { run: 'running' },
            running: { stop: 'idle' }
          }
        });
  
        connect().with('A').mapOnce(A => {
          mapping(A.state.name);
        });
  
        machine.run();
        machine.stop();
  
        expect(mapping).to.be.calledOnce;
        expect(mapping.firstCall).to.be.calledWith(sinon.match('idle'));
      });
    });
    describe('and we use `mapSilent`', function () {
      it('should fire the mapping only when the machine changes its state', function () {
        const mapping = sinon.spy();
        const machine = Machine.create('A', {
          state: { name: 'idle' },
          transitions: {
            idle: { run: 'running' },
            running: { jump: 'jumping' },
            jumping: { stop: 'idle' }
          }
        });
  
        connect().with('A').mapSilent(A => {
          mapping(A.state.name);
        });
  
        machine.run();
        machine.jump();
  
        expect(mapping).to.be.calledTwice;
        expect(mapping.firstCall).to.be.calledWith(sinon.match('running'));
        expect(mapping.secondCall).to.be.calledWith(sinon.match('jumping'));
      });
    });
    describe('and we pass no mapping function', function () {
      it('should still do the connecting', function () {
        const machine = Machine.create('A', {
          state: { name: 'idle' },
          transitions: {
            idle: { run: 'running' },
            running: { stop: 'idle' }
          }
        });

        connect().with('A').map();

        machine.run();

        expect(machine.state).to.deep.equal({ name: 'running' });
      });
    });
    describe('and we have two mappers', function () {
      it('should call them only if the machine that they are connected transitions', function () {
        const machineA = Machine.create('A', {
          state: { name: 'idle' },
          transitions: {
            idle: { run: 'running' },
            running: { stop: 'idle' }
          }
        });
        const machineB = Machine.create('B', {
          state: { name: 'idle' },
          transitions: {
            idle: { run: 'running' },
            running: { stop: 'idle' }
          }
        });
        const spyA = sinon.spy();
        const spyB = sinon.spy();

        connect().with('A').mapSilent(spyA);
        connect().with('B').mapSilent(spyB);

        machineA.run();

        expect(spyA).to.be.calledOnce;
        expect(spyB).to.not.be.called;
      });
    });
    describe('and we have a middleware attached', function () {
      it('should call the proper middleware hook', function () {
        const onMachineConnected = sinon.spy();

        Machine.addMiddleware({ onMachineConnected });
        Machine.create('A', {
          state: { name: 'idle' },
          transitions: {
            idle: { run: 'running' },
            running: { stop: 'idle' }
          }
        });
        Machine.create('B', {
          state: { name: 'waiting' },
          transitions: {
            waiting: { fetch: 'fetching' },
            fetching: { done: 'waiting' }
          }
        });
        connect()
          .with('A', 'B')
          .map(() => {});

        expect(onMachineConnected).to.be.calledOnce.and.to.be.calledWith(sinon.match([
          sinon.match({ name: 'A' }),
          sinon.match({ name: 'B' })
        ]));
      });
    });
    describe('and we destroy a machine', function () {
      it('should remove the machine from any mapping', function () {
        const A = Machine.create('A', {
          state: { name: 'idle' },
          transitions: {
            idle: { run: 'running' },
            running: { stop: 'idle' }
          }
        });
        const B = Machine.create('B', {
          state: { name: 'waiting' },
          transitions: {
            waiting: { fetch: 'fetching' },
            fetching: { done: 'waiting' }
          }
        });
        connect().with('A', 'B').map((A, B) => {});
        connect().with('B').map(B => {});
        connect().with('A').map(A => {});
        
        const [ m1, m2, m3 ] = Object.keys(getMapping());

        expect(getMapping()[m1].machines.map(m => m.name)).to.deep.equal(['A', 'B']);
        expect(getMapping()[m2].machines.map(m => m.name)).to.deep.equal(['B']);
        expect(getMapping()[m3].machines.map(m => m.name)).to.deep.equal(['A']);
        A.destroy();
        expect(getMapping()[m1].machines.map(m => m.name)).to.deep.equal(['B']);
        expect(getMapping()[m2].machines.map(m => m.name)).to.deep.equal(['B']);
        expect(typeof getMapping()[m3]).to.be.equal('undefined');
        B.destroy();
        expect(getMapping()).to.be.deep.equal({});
      });
    });
  });
  describe('when we use the `disconnect` function', function () {
    it('should detach the mapping', function () {
      const mapping = sinon.spy();
      const machine = Machine.create('A', {
        state: { name: 'idle' },
        transitions: {
          idle: { run: 'running' },
          running: { stop: 'idle' }
        }
      });
      const disconnect = connect().with('A').map(A => {
        mapping(A.state.name);
      });

      machine.run();
      disconnect();
      machine.stop();

      expect(mapping).to.be.calledTwice;
      expect(mapping.firstCall).to.be.calledWith(sinon.match('idle'));
      expect(mapping.secondCall).to.be.calledWith(sinon.match('running'));
    });
    describe('and we have a middleware attached', function () {
      it('should call the proper middleware hook', function () {
        const onMachineDisconnected = sinon.spy();

        Machine.addMiddleware({ onMachineDisconnected });
        Machine.create('A', {
          state: { name: 'idle' },
          transitions: {
            idle: { run: 'running' },
            running: { stop: 'idle' }
          }
        });
        Machine.create('B', {
          state: { name: 'waiting' },
          transitions: {
            waiting: { fetch: 'fetching' },
            fetching: { done: 'waiting' }
          }
        });
        connect()
          .with('A', 'B')
          .map(() => {})();

        expect(onMachineDisconnected).to.be.calledOnce.and.to.be.calledWith(sinon.match([
          sinon.match({ name: 'A' }),
          sinon.match({ name: 'B' })
        ]));
      });
    });
  });
});